#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`tutorial-graph`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

# pylint: disable=too-few-public-methods

from dataclasses import (
    dataclass,
)

from ..parse.node import (
    EhnParsePlaceholderBase,
)

################################################################################################################################

@dataclass
class EhnStandardGraph:
    """The standard E-HowNet graph."""

    nodes: dict
    edges: list
    functions: list
    restrictions: list
    root_id: int

class EhnStandardGraphBuilder:
    """The standard E-HowNet graph builder."""

    def __init__(self):
        pass

    def __call__(self, root):
        worker = EhnStandardGraphBuilderWorker(root)
        return EhnStandardGraph(
            nodes=worker.nodes,
            edges=worker.edges,
            functions=worker.functions,
            restrictions=worker.restrictions,
            root_id=worker.root_id,
        )

class EhnStandardGraphBuilderWorker:
    """The standard E-HowNet graph builder worker."""

    def __init__(self, root):
        self.nodes = {}
        self.edges = []
        self.functions = []
        self.restrictions = []

        self.root_id = None

        self.expand_entity(root, is_root=True)

    def expand_entity(self, node, is_root=False):

        node_id = self.get_id(node)
        if is_root:
            self.root_id = node_id
        self.nodes.setdefault(node_id, []).append(node)

        # Placeholder
        if isinstance(node, EhnParsePlaceholderBase):
            if node.value is not None:
                tail_id = self.expand_entity(node.value)
            else:
                tail_id = None
            self.restrictions.append((node_id, tail_id,))

        # Entity
        else:
            # Function Head
            if node.get_function():
                for argument in node.function.arguments:
                    tail_id = self.expand_entity(argument)
                    self.functions.append((node_id, tail_id,))

            # Features
            for feature in node.get_features():
                feature_id = self.expand_feature(feature)
                tail_id = self.expand_entity(feature.value)
                self.edges.append((node_id, feature_id, tail_id,))

        return node_id

    def expand_feature(self, node):
        node_id = self.get_id(node)
        self.nodes.setdefault(node_id, []).append(node)

        # Function Head
        if node.get_function():
            for argument in node.function.arguments:
                tail_id = self.expand_entity(argument)
                self.functions.append((node_id, tail_id,))

        return node_id

    @staticmethod
    def get_id(node):
        return node.get_coindex() or id(node)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`tutorial-db`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

import os
import sqlite3
import warnings

from collections import (
    defaultdict,
)

from treelib import (
    Tree,
)

from .data import (
    EhnDbNode,
    EhnDbNodeData,
    EhnDbNodeType,
    EhnDbWordData,
)

################################################################################################################################

class EhnDb:
    """E-HowNet Database."""

    def __init__(self, *, db_file=None):

        assert db_file is not None, \
            'Please download the database file manually from https://ckip.iis.sinica.edu.tw/CKIP/ehownet_reg/'
        assert os.path.isfile(db_file), f'{db_file} is not a file!'

        self.tree = Tree(node_class=EhnDbNode)

        self.text2nid_concept = defaultdict(list)
        self.text2nid_word = defaultdict(list)
        self.text2nid_partial = defaultdict(list)

        # Load Database
        lite_db = sqlite3.connect(db_file)
        self._load_db(lite_db.cursor())
        lite_db.close()

        # Normalize key-mappings
        self.text2nid_concept = {key: list(set(value)) for key, value in self.text2nid_concept.items()}
        self.text2nid_word = {key: list(set(value)) for key, value in self.text2nid_word.items()}
        self.text2nid_partial = {key: list(set(value)) for key, value in self.text2nid_partial.items()}

    def get_nids(self, text, *, concept=True, word=True, full_match=False):
        res = []
        if concept:
            res += self.text2nid_concept.get(text, [])
            if not full_match:
                res += self.text2nid_partial.get(text, [])
        if word:
            res += self.text2nid_word.get(text, [])
        return sorted(set(res))

    def get_nodes(self, text, **args):
        nids = self.get_nids(text, **args)
        return list(map(self.tree.__getitem__, nids))

    def _load_db(self, cursor):

        cid2child = defaultdict(list)
        cid2data = dict()

        # Load Concept
        cursor.execute('SELECT `id`, `parent_id`, `label`, `defn`, `is_definite` FROM concept')
        for cid, pid, label, defn, definite in cursor.fetchall():
            if pid is None:
                pid = 0
            cid2child[pid].append(cid)
            cid2data[cid] = (label, defn, definite, pid)

        # Build Concept
        self.tree.create_node(tag='ROOT', identifier=0)
        def _build_concept(pid):
            for cid in cid2child[pid]:
                label, defn, definite, _ = cid2data.pop(cid)

                # Create node
                self.tree.create_node(
                    tag=label,
                    identifier=cid,
                    parent=pid, data=EhnDbNodeData(
                        type=EhnDbNodeType.C,
                        defn=defn,
                        definite=bool(definite),
                    ),
                )

                # Register key mapping
                self.text2nid_concept[label].append(cid)
                for sublabel in label.split('|'):
                    if sublabel != label:
                        self.text2nid_partial[sublabel].append(cid)

                # Recursion
                _build_concept(cid)

        _build_concept(0)

        # Check tree
        for cid, (label, _, _, pid,) in cid2data.items():
            warnings.warn(f'Invalid parent ID #{pid}! of Concept#{cid} {label}!')

        # Load Word
        defn2words = defaultdict(list)
        cursor.execute('SELECT `id`, `parent_id`, `label`, `sense_no`, `defn`, `is_definite`, `is_attached` FROM word')
        for wid, pid, label, sense_no, defn, definite, is_attached in cursor.fetchall():
            if pid not in self.tree:
                warnings.warn(f'Invalid parent ID #{pid}! of Word#{wid} {label}#{sense_no}!')
                continue

            if is_attached:
                self.tree[pid].words.append(EhnDbWordData(
                    word=label,
                    sense_no=sense_no,
                ))
                self.text2nid_word[label].append(pid)

            else:
                defn2words[pid, defn].append((wid, label, sense_no, definite,))

        # Build Words
        for (pid, defn), words in defn2words.items():
            words.sort()
            wid, label, _, definite0 = words[0]
            nid = -wid
            node = self.tree.create_node(
                tag=label,
                identifier=nid,
                parent=pid, data=EhnDbNodeData(
                    type=EhnDbNodeType.W,
                    defn=defn,
                    definite=bool(definite0),
                ),
            )

            for _, label, sense_no, _ in words:
                node.words.append(EhnDbWordData(
                    word=label,
                    sense_no=sense_no,
                ))
                self.text2nid_word[label].append(nid)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`tutorial-parse_node`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

# pylint: disable=too-few-public-methods

from .base import (
    EhnParseEntityLike,
    EhnParseFeatureBase,
    EhnParseFunctionBase,
    EhnParseSubjectBase,

    EhnParseAnchorBody,
    EhnParseArgumentBody,
    EhnParseFeatureBody,
    EhnParseStrHead,
)

################################################################################################################################
# Subject
#

class EhnParseSubject(EhnParseSubjectBase, EhnParseFeatureBody, EhnParseAnchorBody):
    """E-HowNet Parsing: Subject Node"""

    node_type = 'Subject'
    feature_type = EhnParseFeatureBase

    def __init__(self, *features):
        EhnParseSubjectBase.__init__(self)
        EhnParseFeatureBody.__init__(self, *features)
        EhnParseAnchorBody.__init__(self, coindex='x?')

    @property
    def head(self):
        return 'SUBJECT'

    def children(self):
        yield from self.features

    def dumps(self):
        return ','.join(feature.dumps() for feature in self.features) if self.features else ''

################################################################################################################################
# Function
#

class EhnParseFunction(EhnParseFunctionBase, EhnParseArgumentBody, EhnParseStrHead):
    """E-HowNet Parsing: Function Node"""

    node_type = 'Function'
    argument_type = EhnParseEntityLike

    def __init__(self, head, *arguments):
        EhnParseFunctionBase.__init__(self)
        EhnParseArgumentBody.__init__(self, *arguments)
        EhnParseStrHead.__init__(self, head)

    def children(self):
        yield from self.arguments

    def dumps(self):
        _arguments = ','.join(argument.dumps() for argument in self.arguments)
        return f'{self.head}({_arguments})'



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`tutorial-graph`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

# pylint: disable=too-few-public-methods

from dataclasses import (
    dataclass,
)

from ..parse.node import (
    EhnParseFunction,
    EhnParseNormalEntity,
    EhnParseRestrictionPlaceholder,
)

################################################################################################################################

@dataclass
class EhnVisGraph:
    """The E-HowNet graph for vis.js."""

    nodes: dict
    edges: list

class EhnVisGraphBuilder:
    """The E-HowNet graph builder for vis.js."""

    def __init__(self, definite_labels=None):
        if definite_labels is not None:
            self.definite_labels = definite_labels
        else:
            self.definite_labels = set()

    def __call__(self, root, label):
        worker = EhnVisGraphBuilderWorker(root, label, definite_labels=self.definite_labels)
        return EhnVisGraph(
            nodes=worker.nodes,
            edges=worker.edges,
        )

class EhnVisGraphBuilderWorker:
    """The E-HowNet graph builder worker for vis.js."""

    def __init__(self, root, label, *, definite_labels):
        self.definite_labels = definite_labels

        self.nodes = {}
        self.edges = []

        tail_id = self.expand_node(root)

        node_id = 'root'
        self.nodes[node_id] = {
            'id': node_id,
            'label': label,
            'group': 'Root'
        }
        self.edges.append({
            'from': node_id,
            'to': tail_id,
            'label': 'IS',
        })

    def expand_node(self, node):
        node_id = node.get_coindex() or str(id(node))

        # Check definite
        is_definite = False
        if isinstance(node, EhnParseNormalEntity) and node.head in self.definite_labels:
            is_definite = True
            node_id = node.head

        # Check function
        is_function = isinstance(node, EhnParseFunction)

        # Restriction
        if isinstance(node, EhnParseRestrictionPlaceholder):

            # Create node
            if node_id not in self.nodes:
                self.nodes[node_id] = {
                    'id': node_id,
                    'label': 'ANY',
                    'group': node.node_type,
                }

            tail_id = self.expand_node(node.value)
            self.edges.append({
                'from': node_id,
                'to': tail_id,
                'label': 'RESTRICT',
            })

        else:

            # Function head
            if node.get_function():
                node_id = self.expand_node(node.function)

            # Create node
            if node_id not in self.nodes:
                self.nodes[node_id] = {
                    'id': node_id,
                    'label': node.head + ('()' if is_function else ''),
                    'group': 'Definite' if is_definite else node.node_type,
                }

            # Features
            for feature in node.get_features():
                tail_id, label = self.expand_feature(feature)
                self.edges.append({
                    'from': node_id,
                    'to': tail_id,
                    'label': label,
                })

            # Get Argument
            for argument in node.get_arguments():
                tail_id = self.expand_node(argument)
                self.edges.append({
                    'from': node_id,
                    'to': tail_id,
                    'label': 'ARGUMENT',
                })

            # Get Value
            if node.get_value():
                tail_id = self.expand_node(node.value)
                self.edges.append({
                    'from': node_id,
                    'to': tail_id,
                    'label': 'VALUE',
                })

        return node_id

    def expand_feature(self, feature):
        if feature.get_function():
            return self.expand_node(feature), 'FEATURE'
        return self.expand_node(feature.value), feature.head



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`tutorial-graph`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from .standard import (
    EhnStandardGraphBuilder,
)

from .vis import (
    EhnVisGraphBuilder,
)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`tutorial-db`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from enum import (
    Enum,
)

from dataclasses import (
    dataclass,
    field,
)

from typing import (
    List,
)

from treelib import (
    Node,
)

################################################################################################################################

class EhnDbNodeType(Enum):
    """E-HowNet Database Node Type."""

    C = 'C'  #: concept.
    W = 'W'  #: word.

@dataclass
class EhnDbWordData:
    """E-HowNet Database Word Data."""

    word: str  #: the word.
    sense_no: int  #: the sense number.

    def __repr__(self):
        return f'<Word {self.word}#{self.sense_no}>'

@dataclass
class EhnDbNodeData:
    """E-HowNet Database Node Data."""

    type: EhnDbNodeType  #: the node type.
    defn: str = None  #: the definition.
    words: List[EhnDbWordData] = field(default_factory=list)  #: the attached words.
    definite: bool = False  #: whether is an instance of not.

class EhnDbNode(Node):
    """E-HowNet Database Node."""

    data_class = EhnDbNodeData

    @property
    def nid(self):
        return self.identifier

    @property
    def label(self):
        return self.tag

    @property
    def defn(self):
        return self.data.defn

    @property
    def type(self):
        return self.data.type

    @property
    def words(self):
        return self.data.words

    @property
    def definite(self):
        return self.data.definite

    def __repr__(self):
        return f'<Node#{self.nid} "{self.label}">'



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`tutorial-db`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from .core import (
    EhnDb,
)

from .data import (
    EhnDbNode,
    EhnDbNodeData,
    EhnDbNodeType,
    EhnDbWordData,
)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`tutorial-parse_node`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

# pylint: disable=too-few-public-methods

import warnings as _warnings

from .base import (
    EhnParseReferenceBase,
    EhnParseStrHead,
)

################################################################################################################################

class EhnParseCoindexReference(EhnParseReferenceBase, EhnParseStrHead):
    """E-HowNet Parsing: Coindex Reference Node"""

    node_type = 'CoindexReference'

    def __init__(self, head):
        EhnParseReferenceBase.__init__(self)
        EhnParseStrHead.__init__(self, head)

    def children(self):
        return []

    @property
    def _tree_label(self):
        return f'[Reference] ${self.head}'

    def dumps(self):
        return f'{{{self.head}}}'

    def get_coindex(self):
        return self.head

################################################################################################################################

class EhnParseSubjectReference(EhnParseReferenceBase):
    """E-HowNet Parsing: Subject Reference Node"""

    node_type = 'SubjectReference'
    def __init__(self):
        EhnParseReferenceBase.__init__(self)

    @property
    def head(self):
        return '$x?'

    def children(self):
        return []

    @property
    def _tree_label(self):
        return f'[SubjectReference] $x?'

    @staticmethod
    def dumps():
        return '{x?}'

    def get_coindex(self):
        return 'x?'

################################################################################################################################

class EhnParseTildeReference(EhnParseReferenceBase):
    """E-HowNet Parsing: Tilde Reference Node

    .. deprecated:: 0.6

    """

    node_type = 'TildeReference'
    def __init__(self):
        EhnParseReferenceBase.__init__(self)
        _warnings.warn('‘~’ is deprecated.', FutureWarning)

    @property
    def head(self):
        return '~'

    def children(self):
        return []

    @staticmethod
    def dumps():
        return '{~}'



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`tutorial-parse_node`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

# pylint: disable=too-few-public-methods

from .base import (
    EhnParseEntityBase,
    EhnParsePlaceholderBase,

    EhnParseAnchorBody,
    EhnParseValueBody,
)

################################################################################################################################
# Restriction
#

class EhnParseRestrictionPlaceholder(EhnParsePlaceholderBase, EhnParseValueBody, EhnParseAnchorBody):
    """E-HowNet Parsing: Restriction Placeholder Node"""

    node_type = 'Restriction'
    value_type = EhnParseEntityBase

    def __init__(self, value, *, coindex=None, anchor=None):
        EhnParsePlaceholderBase.__init__(self)
        EhnParseValueBody.__init__(self, value)
        EhnParseAnchorBody.__init__(self, coindex=coindex, anchor=anchor)

    @property
    def head(self):
        return self.value.head

    def children(self):
        yield self.value

    def dumps(self):
        return f'/{self.value.dumps()}{self.anchor.dumps()}'

################################################################################################################################
# Any
#

class EhnParseAnyPlaceholder(EhnParsePlaceholderBase):
    """E-HowNet Parsing: Any Placeholder Node"""

    node_type = 'Any'

    def __init__(self):
        EhnParsePlaceholderBase.__init__(self)

    @property
    def head(self):
        return 'ANY'

    @property
    def value(self):
        return None

    def children(self):
        return []

    @staticmethod
    def dumps():
        return '{}'



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`tutorial-parse_node`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

from .base import (
    EhnParseNode,
    EhnParseEntityLike,
    EhnParseEntityBase,
    EhnParseReferenceBase,
    EhnParsePlaceholderBase,
    EhnParseFeatureBase,
    EhnParseFunctionBase,
    EhnParseAnchor,
)

from .entity import (
    EhnParseNormalEntity,
    EhnParseFunctionEntity,
    EhnParseNameEntity,
    EhnParseNumberEntity,
)

from .reference import (
    EhnParseCoindexReference,
    EhnParseSubjectReference,
    EhnParseTildeReference,
)

from .placeholder import (
    EhnParseRestrictionPlaceholder,
    EhnParseAnyPlaceholder,
)

from .feature import (
    EhnParseNormalFeature,
    EhnParseFunctionFeature,
)

from .other import (
    EhnParseSubject,
    EhnParseFunction,
)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

"""
Please refer the tutorial ":ref:`tutorial-parse_node`".
"""

__author__ = 'Mu Yang <http://muyang.pro>'
__copyright__ = '2018-2020 CKIP Lab'
__license__ = 'GPL-3.0'

# pylint: disable=protected-access
# pylint: disable=too-few-public-methods

from abc import (
    ABCMeta as _ABCMeta,
    abstractmethod as _abstractmethod,
)

from collections.abc import (
    Sequence as _Sequence,
)

from treelib import (
    Tree as _Tree,
)

################################################################################################################################
# Tree
#
class EhnParseTree(_Tree):

    def show(self, *args, data_property='_tree_label', **kwargs):  # pylint: disable=arguments-differ
        """Print the tree structure."""
        super().show(*args, data_property=data_property, **kwargs)

    def to_str(self, *args, data_property='_tree_label', **kwargs):
        def write(line):
            nonlocal ret
            ret += line.decode() + '\n'
        ret = ''
        self._Tree__print_backend(*args, data_property=data_property, **kwargs, func=write)  # pylint: disable=no-member
        return ret

################################################################################################################################
# Base
#

class EhnParseNode(metaclass=_ABCMeta):
    """E-HowNet Parsing: Base Node"""

    def __init__(self):
        self.__tree = None

    @property
    @_abstractmethod
    def node_type(self):
        return NotImplemented

    @_abstractmethod
    def dumps(self):
        return NotImplemented

    @property
    def _tree_label(self):
        _anchor = f' ${self.anchor.head}' if hasattr(self, 'anchor') and self.anchor.head else ''  # pylint: disable=no-member
        _text = f' {self._tree_label_text}' if hasattr(self, '_tree_label_text') else ''  # pylint: disable=no-member
        return f'[{self.node_type}{_anchor}]{_text}'

    def __str__(self):
        return f'<{self.__class__.__name__} {self.head}>'  # pylint: disable=no-member

    def __repr__(self):
        return str(self)

    #################################################################################

    @_abstractmethod
    def children(self):
        return NotImplemented

    def descendant(self):
        yield self
        for child in self.children():
            yield from child.descendant()

    def tree(self):
        tree = EhnParseTree()
        self._create_tree(tree, None)
        return tree

    def _create_tree(self, tree, parent):
        idx = tree.create_node(parent=parent, data=self).identifier
        for child in self.children():
            child._create_tree(tree, idx)

    #################################################################################

    def get_features(self, key=None):
        res = getattr(self, 'features', [])
        return res if not key else [feature for feature in res if feature.head == key]

    def get_arguments(self):
        return getattr(self, 'arguments', [])

    def get_value(self):
        return getattr(self, 'value', None)

    def get_function(self):
        return getattr(self, 'function', None)

    def get_anchor(self):
        return getattr(self, 'anchor', None)

    def get_coindex(self):
        return getattr(self.get_anchor(), 'head', None)

################################################################################################################################
# Base Types
#

class EhnParseEntityLike(EhnParseNode):  # pylint: disable=abstract-method
    """E-HowNet Parsing: Entity Like Node"""

class EhnParseEntityBase(EhnParseEntityLike):  # pylint: disable=abstract-method
    """E-HowNet Parsing: Base Entity Node"""

class EhnParseReferenceBase(EhnParseEntityLike):  # pylint: disable=abstract-method
    """E-HowNet Parsing: Base Reference Node"""

class EhnParsePlaceholderBase(EhnParseEntityLike):  # pylint: disable=abstract-method
    """E-HowNet Parsing: Base Placeholder Node"""

class EhnParseFeatureBase(EhnParseNode):  # pylint: disable=abstract-method
    """E-HowNet Parsing: Base Feature Node"""

class EhnParseFunctionBase(EhnParseNode):  # pylint: disable=abstract-method
    """E-HowNet Parsing: Base Function Node"""

class EhnParseSubjectBase(EhnParseNode):  # pylint: disable=abstract-method
    """E-HowNet Parsing: Base Subject Node"""

################################################################################################################################
# Anchor
#

class EhnParseAnchor:
    """E-HowNet Parsing: Node Anchor"""

    def __init__(self, head=None):
        self.head = head

    def __str__(self):
        return self.head if self.head else ''

    def __repr__(self):
        return str(self)

    def dumps(self):
        return f'_{self.head}' if self.head else ''

################################################################################################################################
# Heads
#

class EhnParseStrHead(metaclass=_ABCMeta):
    """E-HowNet Parsing: Base Node with String Head"""

    def __init__(self, head):
        self.head = head

    @property
    def head(self):
        return self._head

    @head.setter
    def head(self, head):
        assert isinstance(head, str), f'‘{head}’ is not a str!'
        self._head = head  # pylint: disable=attribute-defined-outside-init

    @property
    def _tree_label_text(self):
        return self.head

################################################################

class EhnParseFunctionHead(metaclass=_ABCMeta):
    """E-HowNet Parsing: Base Node with Function Head"""

    def __init__(self, function):
        self.function = function

    @property
    def head(self):
        return self.function.head

    @property
    def function(self):
        return self._function

    @function.setter
    def function(self, function):
        assert isinstance(function, EhnParseFunctionBase), f'‘{function}’ is not a Function!'
        self._function = function  # pylint: disable=attribute-defined-outside-init

################################################################################################################################
# Bodies
#

class EhnParseValueBody(metaclass=_ABCMeta):
    """E-HowNet Parsing: Base Node with Value"""

    @property
    @_abstractmethod
    def value_type(self):
        return NotImplemented

    def __init__(self, value):
        self.value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        assert isinstance(value, self.value_type), f'‘{value}’ is not a {self.value_type}!'
        self._value = value  # pylint: disable=attribute-defined-outside-init

class EhnParseFeatureBody(_Sequence, metaclass=_ABCMeta):
    """E-HowNet Parsing: Base Node with Feature"""

    @property
    @_abstractmethod
    def feature_type(self):
        return NotImplemented

    def __init__(self, *features):
        self.features = features

    @property
    def features(self):
        return self._features

    @features.setter
    def features(self, features):
        self._features = []  # pylint: disable=attribute-defined-outside-init
        for feature in features:
            self.add_feature(feature)

    def add_feature(self, feature):
        assert isinstance(feature, self.feature_type), f'‘{feature}’ is not a {self.feature_type}!'
        self._features.append(feature)

    def __getitem__(self, key):
        return self._features[key]

    def __len__(self):
        return len(self._features)

class EhnParseArgumentBody(_Sequence, metaclass=_ABCMeta):
    """E-HowNet Parsing: Base Node with Argument"""

    @property
    @_abstractmethod
    def argument_type(self):
        return NotImplemented

    def __init__(self, *arguments):
        self.arguments = arguments

    @property
    def arguments(self):
        return self._arguments

    @arguments.setter
    def arguments(self, arguments):
        self._arguments = []  # pylint: disable=attribute-defined-outside-init
        for argument in arguments:
            self.add_argument(argument)

    def add_argument(self, argument):
        assert isinstance(argument, self.argument_type), f'‘{argument}’ is not a {self.argument_type}!'
        self._arguments.append(argument)

    def __getitem__(self, key):
        return self._arguments[key]

    def __len__(self):
        return len(self._arguments)

################################################################

class EhnParseAnchorBody(metaclass=_ABCMeta):
    """E-HowNet Parsing: Base Node with Anchor"""

    def __init__(self, *, coindex=None, anchor=None):
        self.anchor = anchor or EhnParseAnchor(coindex)

    @property
    def anchor(self):
        return self._anchor

    @anchor.setter
    def anchor(self, anchor):
        assert isinstance(anchor, EhnParseAnchor), f'‘{anchor}’ is not an Anchor!'
        self._anchor = anchor  # pylint: disable=attribute-defined-outside-init




# parsetab.py
# This file is automatically generated. Do not edit.
# pylint: disable=W,C,R
_tabversion = '3.10'

_lr_method = 'LALR'

_lr_signature = 'COINDEX COINDEX0 COLON COMMA EQUAL LBRACE LPAREN NUMBER QUOTE RBRACE RPAREN SLASH TEXT TILDE ULINEexpr : entity\n                | subjectsubject : feature\n                   | subject COMMA featureentity : LBRACE NUMBER RBRACEentity : LBRACE QUOTE TEXT QUOTE RBRACEentityOpen : LBRACE TEXTentityOpen : LBRACE functionentityAnchor : entityOpen anchorentityFeature : entityOpen   COLON feature\n                         | entityAnchor COLON featureentityFeature : entityFeature COMMA featureentity : entityOpen    RBRACE\n                  | entityAnchor  RBRACE\n                  | entityFeature RBRACEreference : LBRACE COINDEX RBRACEreference : LBRACE COINDEX0 RBRACEreference : LBRACE TILDE RBRACErestriction : SLASH entity\n                       | SLASH referencerestriction : SLASH entity anchor\n                       | SLASH reference anchorany : LBRACE RBRACEfeature : TEXT EQUAL entity\n                   | TEXT EQUAL reference\n                   | TEXT EQUAL restriction\n                   | TEXT EQUAL anyfeature : function EQUAL entity\n                   | function EQUAL reference\n                   | function EQUAL restriction\n                   | function EQUAL anyfunction : TEXT LPAREN RPARENfunction : TEXT LPAREN restriction RPARENfunctionOpen : TEXT LPAREN entity\n                        | TEXT LPAREN referencefunctionArgument : functionOpen     COMMA entity\n                            | functionOpen     COMMA reference\n                            | functionArgument COMMA entity\n                            | functionArgument COMMA referencefunction : functionOpen     RPAREN\n                    | functionArgument RPARENanchor : ULINE COINDEX'
    
_lr_action_items = {'LBRACE':([0,18,19,28,30,32,41,],[4,40,46,40,46,46,46,]),'TEXT':([0,4,13,15,22,25,27,40,46,],[5,16,5,35,5,5,5,16,16,]),'$end':([1,2,3,9,20,24,26,33,34,36,37,38,39,48,51,52,53,54,60,64,65,67,68,69,70,71,72,],[0,-1,-2,-3,-13,-14,-15,-4,-5,-24,-25,-26,-27,-42,-28,-29,-30,-31,-23,-19,-20,-6,-16,-17,-18,-21,-22,]),'COMMA':([3,8,9,11,12,20,24,26,33,34,36,37,38,39,44,45,47,48,49,50,51,52,53,54,55,56,57,58,60,64,65,67,68,69,70,71,72,],[13,27,-3,30,32,-13,-14,-15,-4,-5,-24,-25,-26,-27,-34,-35,-10,-42,-11,-12,-28,-29,-30,-31,-36,-37,-38,-39,-23,-19,-20,-6,-16,-17,-18,-21,-22,]),'NUMBER':([4,40,46,],[14,14,14,]),'QUOTE':([4,35,40,46,],[15,59,15,15,]),'EQUAL':([5,10,29,31,42,66,],[18,28,-40,-41,-32,-33,]),'LPAREN':([5,16,],[19,19,]),'RBRACE':([6,7,8,14,16,17,20,21,24,26,29,31,34,36,37,38,39,40,42,47,48,49,50,51,52,53,54,59,60,61,62,63,64,65,66,67,68,69,70,71,72,],[20,24,26,34,-7,-8,-13,-9,-14,-15,-40,-41,-5,-24,-25,-26,-27,60,-32,-10,-42,-11,-12,-28,-29,-30,-31,67,-23,68,69,70,-19,-20,-33,-6,-16,-17,-18,-21,-22,]),'COLON':([6,7,16,17,21,29,31,42,48,66,],[22,25,-7,-8,-9,-40,-41,-32,-42,-33,]),'ULINE':([6,16,17,20,24,26,29,31,34,42,64,65,66,67,68,69,70,],[23,-7,-8,-13,-14,-15,-40,-41,-5,-32,23,23,-33,-6,-16,-17,-18,]),'RPAREN':([11,12,19,20,24,26,34,43,44,45,48,55,56,57,58,64,65,67,68,69,70,71,72,],[29,31,42,-13,-14,-15,-5,66,-34,-35,-42,-36,-37,-38,-39,-19,-20,-6,-16,-17,-18,-21,-22,]),'SLASH':([18,19,28,],[41,41,41,]),'COINDEX':([23,40,46,],[48,61,61,]),'COINDEX0':([40,46,],[62,62,]),'TILDE':([40,46,],[63,63,]),}

_lr_action = {}
for _k, _v in _lr_action_items.items():
   for _x,_y in zip(_v[0],_v[1]):
      if not _x in _lr_action:  _lr_action[_x] = {}
      _lr_action[_x][_k] = _y
del _lr_action_items

_lr_goto_items = {'expr':([0,],[1,]),'entity':([0,18,19,28,30,32,41,],[2,36,44,51,55,57,64,]),'subject':([0,],[3,]),'entityOpen':([0,18,19,28,30,32,41,],[6,6,6,6,6,6,6,]),'entityAnchor':([0,18,19,28,30,32,41,],[7,7,7,7,7,7,7,]),'entityFeature':([0,18,19,28,30,32,41,],[8,8,8,8,8,8,8,]),'feature':([0,13,22,25,27,],[9,33,47,49,50,]),'function':([0,4,13,22,25,27,40,46,],[10,17,10,10,10,10,17,17,]),'functionOpen':([0,4,13,22,25,27,40,46,],[11,11,11,11,11,11,11,11,]),'functionArgument':([0,4,13,22,25,27,40,46,],[12,12,12,12,12,12,12,12,]),'anchor':([6,64,65,],[21,71,72,]),'reference':([18,19,28,30,32,41,],[37,45,52,56,58,65,]),'restriction':([18,19,28,],[38,43,53,]),'any':([18,28,],[39,54,]),}

_lr_goto = {}
for _k, _v in _lr_goto_items.items():
   for _x, _y in zip(_v[0], _v[1]):
       if not _x in _lr_goto: _lr_goto[_x] = {}
       _lr_goto[_x][_k] = _y
del _lr_goto_items
_lr_productions = [
  ("S' -> expr","S'",1,None,None,None),
  ('expr -> entity','expr',1,'p_expr','parser.py',169),
  ('expr -> subject','expr',1,'p_expr','parser.py',170),
  ('subject -> feature','subject',1,'p_subject','parser.py',175),
  ('subject -> subject COMMA feature','subject',3,'p_subject','parser.py',176),
  ('entity -> LBRACE NUMBER RBRACE','entity',3,'p_entity_number','parser.py',185),
  ('entity -> LBRACE QUOTE TEXT QUOTE RBRACE','entity',5,'p_entity_name','parser.py',189),
  ('entityOpen -> LBRACE TEXT','entityOpen',2,'p_entity_normal_open','parser.py',193),
  ('entityOpen -> LBRACE function','entityOpen',2,'p_entity_function_open','parser.py',197),
  ('entityAnchor -> entityOpen anchor','entityAnchor',2,'p_entity_anchor','parser.py',201),
  ('entityFeature -> entityOpen COLON feature','entityFeature',3,'p_entity_feature0','parser.py',206),
  ('entityFeature -> entityAnchor COLON feature','entityFeature',3,'p_entity_feature0','parser.py',207),
  ('entityFeature -> entityFeature COMMA feature','entityFeature',3,'p_entity_feature','parser.py',212),
  ('entity -> entityOpen RBRACE','entity',2,'p_entity_close','parser.py',217),
  ('entity -> entityAnchor RBRACE','entity',2,'p_entity_close','parser.py',218),
  ('entity -> entityFeature RBRACE','entity',2,'p_entity_close','parser.py',219),
  ('reference -> LBRACE COINDEX RBRACE','reference',3,'p_reference_coindex','parser.py',224),
  ('reference -> LBRACE COINDEX0 RBRACE','reference',3,'p_reference_subject','parser.py',228),
  ('reference -> LBRACE TILDE RBRACE','reference',3,'p_reference_tilde','parser.py',232),
  ('restriction -> SLASH entity','restriction',2,'p_restriction','parser.py',237),
  ('restriction -> SLASH reference','restriction',2,'p_restriction','parser.py',238),
  ('restriction -> SLASH entity anchor','restriction',3,'p_restriction_anchor','parser.py',242),
  ('restriction -> SLASH reference anchor','restriction',3,'p_restriction_anchor','parser.py',243),
  ('any -> LBRACE RBRACE','any',2,'p_any','parser.py',247),
  ('feature -> TEXT EQUAL entity','feature',3,'p_feature','parser.py',252),
  ('feature -> TEXT EQUAL reference','feature',3,'p_feature','parser.py',253),
  ('feature -> TEXT EQUAL restriction','feature',3,'p_feature','parser.py',254),
  ('feature -> TEXT EQUAL any','feature',3,'p_feature','parser.py',255),
  ('feature -> function EQUAL entity','feature',3,'p_function_feature','parser.py',259),
  ('feature -> function EQUAL reference','feature',3,'p_function_feature','parser.py',260),
  ('feature -> function EQUAL restriction','feature',3,'p_function_feature','parser.py',261),
  ('feature -> function EQUAL any','feature',3,'p_function_feature','parser.py',262),
  ('function -> TEXT LPAREN RPAREN','function',3,'p_function_any','parser.py',267),
  ('function -> TEXT LPAREN restriction RPAREN','function',4,'p_function_restriction','parser.py',271),
  ('functionOpen -> TEXT LPAREN entity','functionOpen',3,'p_function_open','parser.py',275),
  ('functionOpen -> TEXT LPAREN reference','functionOpen',3,'p_function_open','parser.py',276),
  ('functionArgument -> functionOpen COMMA entity','functionArgument',3,'p_function_argument','parser.py',280),
  ('functionArgument -> functionOpen COMMA reference','functionArgument',3,'p_function_argument','parser.py',281),
  ('functionArgument -> functionArgument COMMA entity','functionArgument',3,'p_function_argument','parser.py',282),
  ('functionArgument -> functionArgument COMMA reference','functionArgument',3,'p_function_argument','parser.py',283),
  ('function -> functionOpen RPAREN','function',2,'p_function_close','parser.py',288),
  ('function -> functionArgument RPAREN','function',2,'p_function_close','parser.py',289),
  ('anchor -> ULINE COINDEX','anchor',2,'p_anchor','parser.py',294),
]
